(*
    Greenfish Icon Editor Pro
    Copyright (c) 2012-13 B. Szalkai

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)
// Interface to the ImageConverter utility, for loading/saving certain images.
unit ImageConverter_Intf;

{$mode delphi}

interface

uses
  Classes, SysUtils, DocClass;

const
  IMAGE_CONVERTER_EXE = 'ImageConverter\bin\Debug\ImageConverter.exe';
  IMAGE_CONVERTER_DOTNET_VERSION = '3.5';

type

  { TSvgInfo }

  TSvgInfo = record
    Width: double; // default document width, in pixels
    Height: double; // default document height, in pixels
    Dpi: double; // default document dpi

    procedure Clear;
  end;

function ImageConverterAvailable: boolean;
function ImageConverterInstalled: boolean;
function ImageConverterFrameworkInstalled: boolean;

function svgDetect(Data: Pointer; Size: integer): boolean;
function svgGetInfo(var info: TSvgInfo; const fn: string): boolean;
function svgLoadFromStream(Doc: TIconDoc; s: TStream; scaleFactor: double): boolean;

function tiffDetect(Data: Pointer; Size: integer): boolean;
function tiffLoadFromStream(Doc: TIconDoc; s: TStream): boolean;
function tiffSaveToStream(Doc: TIconDoc; s: TStream): TieDataLosses;

implementation

uses
  ieShared, dlgDebug, LangPack;

function ImageConverterInstalled: boolean;
begin
  Result := FileExists(IMAGE_CONVERTER_EXE);
end;

function ImageConverterFrameworkInstalled: boolean;
begin
  Result := DotNetInstalled(IMAGE_CONVERTER_DOTNET_VERSION);
end;

function ImageConverterAvailable: boolean;
begin
  Result := ImageConverterInstalled and ImageConverterFrameworkInstalled;
end;

function anyLoadFromStream(Doc: TIconDoc; s: TStream; srcType, destType, param: string): boolean;
var
  HelperExitCode: integer;
  fnInput, fnOutput: string;
  fs: TFileStream;

begin
  fnInput := GetTempFileName;

  fs := TFileStream.Create(fnInput, fmCreate);
  try
    fs.CopyFrom(s, s.Size - s.Position);
  finally
    fs.Free;
  end;

  fnOutput := GetTempFileName;

  HelperExitCode := RunCommandHidden(AppDir, IMAGE_CONVERTER_EXE,
    [srcType+'2'+destType, fnInput, fnOutput, param, '--silent']);
  if HelperExitCode = 0 then
  begin
    Result := Doc.LoadFromFile(fnOutput, Pref_MaxWidth, Pref_MaxHeight, 1, lpGet('LY_BACKGROUND')) <> iftNone;
  end else
  begin
    {$IFDEF GFIEDEBUG}
      Log('ImageConverter exited with code '+IntToStr(HelperExitCode));
    {$ENDIF}
    Result := False;
  end;

  DeleteFile(fnInput);
  DeleteFile(fnOutput);
end;

function anySaveToStream(Doc: TIconDoc; s: TStream; srcType, destType, param: string): TieDataLosses;
var
  HelperExitCode: integer;
  fnInput, fnOutput: string;
  fs: TFileStream;
  ft: TImageFileType;

begin
  fnInput := GetTempFileName; // a file that GFIE can produce

  if srcType = 'png' then
    ft := iftPng
  else
    ft := iftGfie;
  Doc.SaveToFile(fnInput, ft, 0, '', 0, false, Result);

  fnOutput := GetTempFileName;

  HelperExitCode := RunCommandHidden(AppDir, IMAGE_CONVERTER_EXE,
    [srcType+'2'+destType, fnInput, fnOutput, param, '--silent']);
  if HelperExitCode = 0 then
  begin
    fs := TFileStream.Create(fnOutput, fmOpenRead);
    try
      s.CopyFrom(fs, fs.Size);
    finally
      fs.Free;
    end;
  end
  else
  begin
{$IFDEF GFIEDEBUG}
  Log('ImageConverter exited with code '+IntToStr(HelperExitCode));
{$ENDIF}
    Result := [dlError];
  end;

  DeleteFile(fnInput);
  DeleteFile(fnOutput);
end;

function svgDetect(Data: Pointer; Size: integer): boolean;
var
  sHeader: string;
  DataStart, DataEnd, p: PChar;
begin
  sHeader := '<?xml';
  DataStart := PChar(Data);
  DataEnd := DataStart + Size - Length(sHeader);
  p := DataStart;
  while p <= DataEnd do
  begin
    if strlicomp(@sHeader[1], p, Length(sHeader)) = 0 then
      Exit(True);
    inc(p);
  end;
  Exit(False);
end;

function tiffDetect(Data: Pointer; Size: integer): boolean;
begin
  Result := (Size >= 4) and ((PCardinal(Data)^ = $002a4949) or (PCardinal(Data)^ = $2a004d4d));
end;

function svgGetInfo(var info: TSvgInfo; const fn: string): boolean;
var
  HelperExitCode, p: integer;
  s, key, value, fnOutput: string;
  f: textfile;

begin
  info.Clear;
  fnOutput := GetTempFileName;

  HelperExitCode := RunCommandHidden(AppDir, IMAGE_CONVERTER_EXE,
    ['svginfo', fn, fnOutput, '', '--silent']);
  if HelperExitCode = 0 then
  begin
    Result := True;
    AssignFile(f, fnOutput);
    Reset(f);
    try
      while not Eof(f) do
      begin
        Readln(f, s);
        p := Pos('=', s);
        if p <= 0 then Continue;
        key := Copy(s, 1, p-1);
        value := Copy(s, p+1, Length(s)-p);
        if key = 'width_px' then
          info.Width := StrToFloat(value)
        else if key = 'height_px' then
          info.Height := StrToFloat(value)
        else if key = 'dpi' then
          info.Dpi := StrToFloat(value);
      end;
    finally
      CloseFile(f);
    end;
  end else
  begin
    {$IFDEF GFIEDEBUG}
      Log('ImageConverter svginfo exited with code '+IntToStr(HelperExitCode));
    {$ENDIF}
    Result := False;
  end;

  DeleteFile(fnOutput);
end;

function svgLoadFromStream(Doc: TIconDoc; s: TStream; scaleFactor: double): boolean;
begin
  Result := anyLoadFromStream(Doc, s, 'svg', 'gfie', FloatToStr(scaleFactor));
end;

function tiffLoadFromStream(Doc: TIconDoc; s: TStream): boolean;
begin
  Result := anyLoadFromStream(Doc, s, 'tiff', 'gfie', '');
end;

function tiffSaveToStream(Doc: TIconDoc; s: TStream): TieDataLosses;
var
  DocToSave: TIconDoc;

begin
  if Doc.HasNonTrivialLayers then
  begin
    // TODO do not copy all layers then flatten. Just render them to DocToSave in one pass.
    DocToSave := TIconDoc.Create;
    DocToSave.Assign(Doc);
    DocToSave.Flatten;
  end else
    DocToSave := Doc;

  try
    Result := anySaveToStream(DocToSave, s, 'gfie', 'tiff', '');
  finally
    if DocToSave <> Doc then
      DocToSave.Free;
  end;
end;

{ TSvgInfo }

procedure TSvgInfo.Clear;
begin
  Width := 0;
  Height := 0;
  Dpi := 96;
end;

end.

